from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestMoveElement(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.source = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.destination = cls.corpus.elements.get(name='Volume 2')

    def test_move_element_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': str(self.destination.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_move_element_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': str(self.destination.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_move_element_wrong_acl(self):
        private_corpus = Corpus.objects.create(name='private', public=False)
        private_element = private_corpus.elements.create(
            type=private_corpus.types.create(slug='folder'),
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:move-element'), {'source': str(private_element.id), 'destination': str(private_element.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                'source': [f'Invalid pk "{private_element.id}" - object does not exist.'],
                'destination': [f'Invalid pk "{private_element.id}" - object does not exist.']
            }
        )

    def test_move_element_wrong_source(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:move-element'), {'source': '12341234-1234-1234-1234-123412341234', 'destination': str(self.destination.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'source': ['Invalid pk "12341234-1234-1234-1234-123412341234" - object does not exist.']}
        )

    def test_move_element_wrong_destination(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': '12341234-1234-1234-1234-123412341234'}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'destination': ['Invalid pk "12341234-1234-1234-1234-123412341234" - object does not exist.']}
        )

    def test_move_element_same_source_destination(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': str(self.source.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'destination': ['A source element cannot be moved into itself']}
        )

    def test_move_element_different_corpus(self):
        corpus2 = Corpus.objects.create(name='new')
        corpus2.memberships.create(user=self.user, level=Role.Contributor.value)
        destination = corpus2.elements.create(type=corpus2.types.create(slug='folder'))

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': str(destination.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'destination': ['A source element cannot be moved to a destination from another corpus']}
        )

    def test_move_element_destination_is_direct_parent(self):
        destination = self.corpus.elements.get(name='Volume 1')

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': str(destination.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'destination': [f"'{destination.id}' is already a direct parent of '{self.source.id}'"]}
        )

    def test_move_element_destination_is_child(self):
        source = self.corpus.elements.get(name='Volume 1')
        destination_id = self.source.id

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:move-element'), {'source': str(source.id), 'destination': str(destination_id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'destination': [f"'{destination_id}' is a child of element '{source.id}'"]}
        )

    @patch('arkindex.project.triggers.documents_tasks.move_element.delay')
    def test_move_element(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:move-element'), {'source': str(self.source.id), 'destination': str(self.destination.id)}, format='json')
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            source=self.source,
            destination=self.destination,
            user_id=self.user.id,
            description=f"Moving element {self.source.name} to element {self.destination.name}"
        ))
