import logging

from django.conf import settings
from django.db import connections
from django.db.models import Manager, Q
from django.db.models.functions import Coalesce
from django.utils.functional import cached_property

from arkindex.users.managers import BaseACLManager
from arkindex.users.models import Role

logger = logging.getLogger(__name__)


class ActivityManager(Manager):
    """Model management for worker activities"""

    def bulk_insert(self, worker_version_id, process_id, configuration_id, model_version_id, elements_qs):
        """
        Create initial worker activities from a queryset of elements in a efficient way.
        Due to the possible large amount of elements, we use a bulk insert from the elements query (best performances).
        The `ON CONFLICT` clause allows either skipping existing activities if they are in a `Started` or `Processed`
        state, or stealing them from another process if they are `Queued` or `Error`.
        """
        sql, params = elements_qs.distinct().values('id').query.sql_with_params()
        select_params = (worker_version_id, configuration_id, model_version_id, process_id) + params

        # With ON CONFLICT, the target constraint is only optional when the action is DO NOTHING.
        # For DO UPDATE, the target can either be a constraint by name with `ON CONSTRAINT <name>`,
        # or the definition of an index that PostgreSQL will look for, using a list of columns,
        # a collation, an operator class, a predicate, etc.
        # `ON CONSTRAINT <name>` does not work here because it is unable to detect partial constraints.
        # We therefore have to explicitly define one of the partial unique constraints ourselves.
        conflict_target_fields = ['worker_version_id', 'element_id']
        if configuration_id is None:
            conflict_target_condition = 'configuration_id IS NULL'
        else:
            conflict_target_fields.append('configuration_id')
            conflict_target_condition = 'configuration_id IS NOT NULL'

        if model_version_id is None:
            conflict_target_condition += ' AND model_version_id IS NULL'
        else:
            conflict_target_fields.append('model_version_id')
            conflict_target_condition += ' AND model_version_id IS NOT NULL'

        conflict_target = f'({", ".join(conflict_target_fields)}) WHERE {conflict_target_condition}'

        with connections['default'].cursor() as cursor:
            cursor.execute(
                f"""
                INSERT INTO process_workeractivity
                    (element_id, worker_version_id, configuration_id, model_version_id, state, process_id, id, created, updated)
                SELECT
                    elt.id, %s, %s, %s, 'queued', %s, uuid_generate_v4(), current_timestamp, current_timestamp
                FROM ({sql}) AS elt
                ON CONFLICT {conflict_target} DO UPDATE SET
                    process_id = EXCLUDED.process_id,
                    state = EXCLUDED.state,
                    created = EXCLUDED.created
                WHERE process_workeractivity.state <> 'started' AND process_workeractivity.state <> 'processed'
                """,
                select_params)


class CorpusWorkerVersionManager(Manager):

    def rebuild(self):
        """
        Rebuild the corpus worker versions cache from all ML results.
        For compatibility, results without a worker run are generated from the worker version instead.
        """
        from arkindex.documents.models import (
            Classification,
            Element,
            Entity,
            MetaData,
            Transcription,
            TranscriptionEntity,
        )

        querysets = [
            (
                Element.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce('worker_run__version_id', 'worker_version_id'))
                .values_list('corpus_id', 'version_id', 'worker_run__model_version_id', 'worker_run__configuration_id')
            ), (
                Transcription.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce('worker_run__version_id', 'worker_version_id'))
                .values_list('element__corpus_id', 'version_id', 'worker_run__model_version_id', 'worker_run__configuration_id')
            ), (
                Entity.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce('worker_run__version_id', 'worker_version_id'))
                .values_list('corpus_id', 'version_id', 'worker_run__model_version_id', 'worker_run__configuration_id')
            ), (
                TranscriptionEntity.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce('worker_run__version_id', 'worker_version_id'))
                .values_list('entity__corpus_id', 'version_id', 'worker_run__model_version_id', 'worker_run__configuration_id')
            ), (
                Classification.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce('worker_run__version_id', 'worker_version_id'))
                .values_list('element__corpus_id', 'version_id', 'worker_run__model_version_id', 'worker_run__configuration_id')
            ), (
                MetaData.objects
                .exclude(worker_version_id=None)
                .annotate(version_id=Coalesce('worker_run__version_id', 'worker_version_id'))
                .values_list('element__corpus_id', 'version_id', 'worker_run__model_version_id', 'worker_run__configuration_id')
            ),
        ]

        for i, queryset in enumerate(querysets, start=1):
            logger.info(f'Rebuilding cache from {queryset.model.__name__} ({i}/{len(querysets)})')
            self.bulk_create([
                self.model(
                    corpus_id=corpus_id,
                    worker_version_id=worker_version_id,
                    model_version_id=model_version_id,
                    worker_configuration_id=worker_configuration_id,
                )
                for corpus_id, worker_version_id, model_version_id, worker_configuration_id in queryset.distinct()
            ], ignore_conflicts=True)


class WorkerVersionManager(Manager):

    @cached_property
    def imports_version(self):
        """
        WorkerVersion used for all import tasks.
        """
        return (
            self
            # Required by WorkerRun.build_summary
            .select_related('worker', 'revision')
            .prefetch_related('revision__refs')
            .get(id=settings.IMPORTS_WORKER_VERSION)
        )


class WorkerManager(BaseACLManager):

    def executable(self, user):
        """
        Workers that can be viewed and executed by a user.
        """
        # Without authentication, or as a Ponos agent, nothing is executable
        # Agents are not users, filtering on them will make Django raise exceptions, so we have to handle them separately.
        if user.is_anonymous or getattr(user, 'is_agent', False):
            return self.none()

        # Admins can execute anything
        if user.is_admin:
            return self.all()

        from arkindex.process.models import Repository

        return self.filter(
            # All public workers are executable by everyone
            Q(public=True)
            # Otherwise, require contributor rights on either the worker or the worker's repository
            | Q(id__in=self.filter_rights(user, self.model, Role.Contributor.value).values('id'))
            | Q(repository_id__in=self.filter_rights(user, Repository, Role.Contributor.value).values('id'))
        ).distinct()
