from django.conf import settings
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, path, re_path

from arkindex.metrics.urls import metrics_urls
from arkindex.project.api_v1 import api
from arkindex.project.views import CdnHome, FrontendView, OpenAPIDocsView, RobotsTxt

# Fallback to the dummy frontend view when CDN_ASSETS_URL is not set
frontend_view = FrontendView if settings.CDN_ASSETS_URL is None else CdnHome

urlpatterns = [
    path('api/v1/', include((api, 'api'), namespace='api')),
    path('metrics/', include((metrics_urls, 'metrics'), namespace='metrics')),
    path('api-docs/', OpenAPIDocsView.as_view(), name='openapi-docs'),
    path('admin/', admin.site.urls),
    path('rq/', include('django_rq.urls')),
    # Frontend URLs the backend needs with django.urls.reverse
    # Link sent via email for password resets
    path('user/reset/<uidb64>/<token>/', frontend_view.as_view(), name='password_reset_confirm'),
    # Redirection URL for successful OAuth2 flows
    path('process/credentials/', frontend_view.as_view(), name='credentials'),
    path('process/<uuid:pk>/<int:run>', frontend_view.as_view(), name='frontend-process-details'),
    # Link to the corpus management page, shown in the Django admin
    path('corpus/<uuid:pk>', frontend_view.as_view(), name='frontend-corpus-details'),
]

if settings.ROBOTS_TXT_DISALLOW:
    # Add an URL to serve robots.txt
    urlpatterns.append(path('robots.txt', RobotsTxt.as_view(), name='robots'))

if 'debug_toolbar' in settings.INSTALLED_APPS:
    import debug_toolbar
    urlpatterns = [
        path('__debug__/', include(debug_toolbar.urls)),
    ] + urlpatterns

if settings.DEBUG:
    # Serve static files in development—this is done by manage.py runserver, but not by Daphne in dev
    urlpatterns.extend(staticfiles_urlpatterns())

# Add index.html using CDN assets
# It's served as a full fallback on 404 to support loading specific frontend urls
# The (?!api/) in the regex makes sure that any URL starting in api/ will not match this regex, so it will cause a 404,
# but everything else redirects to index.html; this avoids getting the frontend as a response to an incorrect API URL.
if settings.CDN_ASSETS_URL is not None:
    urlpatterns.append(re_path(r'^(?!api/).*$', CdnHome.as_view(), name="home"))
